/*
 * (Super)Framebuffer driver for VESA compliant graphic boards
 *
 * Version 1.2
 *
 * (C) 2003 Sergio Costas Rodriguez (raster@rastersoft.com)
 *
 * http://www.rastersoft.com
 *
 * This driver allows to change the graphic mode using a daemon (fbsetmoded)
 *
 * It's heavily based in VESA 2.0 FB driver
 * (which is (C) 1998 Gerd Knorr <kraxel@goldbach.in-berlin.de>)
 * That's why switching to graphics mode happens at boot time
 * (while running in real mode)
 *
 * This driver needs the FBSETMODED daemon, which uses the great
 * LRMI library (http://sourceforge.net/projects/rlmi)
 * 
 * This driver is released under the GPL v2 license or later.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/console.h>
#include <linux/selection.h>
#include <linux/ioport.h>
#include <linux/init.h>

#include <asm/io.h>
#include <asm/mtrr.h>
#include <asm/uaccess.h>

#include <video/fbcon.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb24.h>
#include <video/fbcon-cfb32.h>
#include <video/fbcon-mac.h>

#include <linux/kmod.h> // we need it for launching the V86 daemon

#include <linux/proc_fs.h> // we need it to be able to communicate with de daemon

#include <linux/sched.h> // we need it to have spinlocks

#define dac_reg	(0x3c8)
#define dac_val	(0x3c9)

/* --------------------------------------------------------------------- */

struct proc_dir_entry *vesafb_fichero_proc; // this is the /PROC file descriptor for communication
struct proc_dir_entry *vesafb_specs_proc; // this is the /PROC file descriptor for supported modes

static volatile unsigned char vesafb_daemon_command[7]; // contains the command to send to the daemon

static volatile spinlock_t vesafb_varlock=SPIN_LOCK_UNLOCKED; // to manage the access to the command and launched variables

/* to manage the launch of the daemon

	bit 0: has to change the graphic mode
	bit 1: has to change a palette entry
	bit 2: there's a daemon launched at this moment */

static volatile unsigned char vesafb_dflags = 0;

/* if 1, the daemon sent us the capabilities, so we can now change modes.
   if 0, the daemon has'n send us the capabilities, so we must work like the old VESAFB driver
   if 2, the daemon sent us the capabilites, but the default mode isn't in that list
*/

static volatile unsigned char vesafb_got_capabilities = 0;

/* palette with colors to send to the daemon.
   order: RGB
   since each entry has 6 bits, we use the bit 7 from red to know if
   we have to send that entry */

static unsigned char vesafb_palete[768];

/* vesafb_color_data stores de size and offsett for each colour in each mode

vesafb_color_data[mode][color][data]

mode: 0=8bpp, 1=15bpp, 2=16bpp, 3=24bpp, 4=32bpp
color: 0=Red, 1=Green, 2=Blue
data: 0=size, 1=offsett

*/

unsigned char vesafb_color_data[5][3][2];

/* supported_modes are the modes supported by the card:

	+----+----+----+----+----+----+
	|  res X  |  res Y  |Bpp |Mode|
	+----+----+----+----+----+----+
	
	The list ends with an entry filled with 0s
	
	vesafb_supported_modes2 are a temporary space where we store the boot mode,
	used until the daemon sends all the available modes

*/

static unsigned short int vesafb_supported_modes2[6];
static unsigned short int *vesafb_supported_modes;

/* --------------------------------------------------------------------- */

/*
 * card parameters
 */

/* card */
static unsigned long video_base; /* physical addr */
static int   video_size;
static char *video_vbase;        /* mapped */

/* mode */
static int  video_bpp;
static int  video_width;
static int  video_height;
static int  video_height_virtual;
static int  video_type = FB_TYPE_PACKED_PIXELS;
static int  video_visual;
static int  video_linelength;
static int  video_cmap_len;

/* --------------------------------------------------------------------- */

static struct fb_var_screeninfo vesafb_defined = {
    0,0,0,0,	/* W,H, W, H (virtual) load xres,xres_virtual*/
    0,0,		/* virtual -> visible no offset */
    8,		/* depth -> load bits_per_pixel */
    0,		/* greyscale ? */
    {0,0,0},	/* R */
    {0,0,0},	/* G */
    {0,0,0},	/* B */
    {0,0,0},	/* transparency */
    0,		/* standard pixel format */
    FB_ACTIVATE_NOW,
    -1,-1,
    0,
    0L,0L,0L,0L,0L,
    0L,0L,0,	/* No sync info */
    FB_VMODE_NONINTERLACED,
    {0,0,0,0,0,0}
};

static struct display disp;
static struct fb_info fb_info;
static struct { u_short blue, green, red, pad; } palette[256];
static int allow_set=0;
static union {
    u16 cfb16[16];
    u32 cfb24[16];
    u32 cfb32[16];
} fbcon_cmap;

static int             inverse   = 0;
#ifdef __x86_64__
static int             mtrr      = 1;
#else
static int             mtrr      = 0;
#endif

static int             currcon   = 0;

static int             pmi_setpal = 0;	/* pmi for palette changes ??? */
static int             ypan       = 0;  /* 0..nothing, 1..ypan, 2..ywrap */
static unsigned short  *pmi_base  = 0;
static void            (*pmi_start)(void);
static void            (*pmi_pal)(void);

static struct display_switch vesafb_sw;

/* --------------------------------------------------------------------- */

static int vesafb_read_specs(char *page,char **start,off_t off,int count,int *eof,void *data) {
	
	unsigned char cadena[30];
	unsigned char *mpointer;
	unsigned int count1,len,bucle,end;
	unsigned short int *memory=NULL;
	unsigned short int fresx,fresy,fbpp;
	
	count1=0;
	mpointer=page;	
	
	switch(vesafb_got_capabilities) {
	case 0:
		memory=vesafb_supported_modes2;
		*(mpointer++)='*'; // to indicate that the daemon hasn't send us the modes
		count1++;
	break;
	case 1:
		memory=vesafb_supported_modes; // the daemon sent us the modes, so display them
	break;
	case 2:
		memory=vesafb_supported_modes;
		// *(mpointer++)='!'; // to indicate that the daemon sent the modes but the default mode isn't there
		// count1++;
	break;
	}
		
	end=1;
	len=0;
	
	do {
		fresx=*(memory++);
		fresy=*(memory++);
		fbpp=((*memory) >> 8) &	0x00FF;
		memory++;
		if (fresx!=0) {
			
			switch(vesafb_got_capabilities) {
			case 0:
			case 1:
				len=sprintf(cadena,"%dx%d (%d bpp)\n",fresx,fresy,fbpp);
				break;
			case 2:
				len=sprintf(cadena,"!%dx%d (%d bpp)\n",fresx,fresy,fbpp);
				break;
			}
			
			for (bucle=0;bucle<len;bucle++) {
				if((count1>=off)&&(count1<(off+count)))
					*(mpointer++)=cadena[bucle];
				if(count1>=(off+count))
					end=0;
				count1++;
			}
		}
	} while(fresx != 0);
	
	*eof=end;
	if((count1-off)>=0)
		return (count1-off);
	else
		return 0;
}

static int vesafb_find_mode(unsigned short int resx, unsigned short int resy,
							unsigned short int bpp, unsigned short int *mode) {

	unsigned short int *memory;

	unsigned short int fresx,fresy,fbpp,fmode;
	int found;

	if(vesafb_got_capabilities == 1) // modes are available
		memory=vesafb_supported_modes;
	else
		memory=vesafb_supported_modes2; // test if the mode to find is the current mode
	
	found=0;
	
	do {
		fresx=*(memory++);
		fresy=*(memory++);
		fbpp=((*memory) >> 8) &	0x00FF;
		fmode=(*memory) & 0x00FF;
		memory++;
		if ((fresx==resx)&&(fresy==resy)&&(fbpp==bpp))
			found=1;
	} while((fresx != 0)&&(found == 0));
	
	*mode=fmode;
	return (found);
}


static int vesafb_pan_display(struct fb_var_screeninfo *var, int con,
                              struct fb_info *info)
{
#ifdef __i386__
    int offset;
    
    if (!ypan)
	return -EINVAL;
    if (var->xoffset)
	return -EINVAL;
    if (var->yoffset > var->yres_virtual)
	return -EINVAL;
    if ((ypan==1) && var->yoffset+var->yres > var->yres_virtual)
	return -EINVAL;
    
    offset = (var->yoffset * video_linelength + var->xoffset) / 4;
    
    __asm__ __volatile__(
			 "call *(%%edi)"
			 : /* no return value */
			 : "a" (0x4f07),         /* EAX */
			 "b" (0),              /* EBX */
			 "c" (offset),         /* ECX */
			 "d" (offset >> 16),   /* EDX */
			 "D" (&pmi_start));    /* EDI */
			
#endif
	
    return 0;
}

static int vesafb_update_var(int con, struct fb_info *info)
{
    if (con == currcon && ypan) {
		struct fb_var_screeninfo *var = &fb_display[currcon].var;
		return vesafb_pan_display(var,con,info);
    }
    return 0;
}

static int vesafb_get_fix(struct fb_fix_screeninfo *fix, int con,
			     struct fb_info *info)
{
    memset(fix, 0, sizeof(struct fb_fix_screeninfo));
    strcpy(fix->id,"VESA VGA");
    
    fix->smem_start=video_base;
    fix->smem_len=video_size;
    fix->type = video_type;
    fix->xpanstep  = 0;
    fix->ypanstep  = ypan     ? 1 : 0;
    fix->ywrapstep = (ypan>1) ? 1 : 0;
	
	if(con==-1) {
		fix->visual = video_visual;
    	fix->line_length=video_linelength;
	} else {
		fix->visual = fb_display[con].visual;
    	fix->line_length = fb_display[con].line_length;
	}
		
    return 0;
}

static int vesafb_get_var(struct fb_var_screeninfo *var, int con,
			     struct fb_info *info) {

    unsigned int temporal;

    if(con==-1) {
		memcpy(var, &vesafb_defined, sizeof(struct fb_var_screeninfo));
		var->xres=video_width;
		var->yres=video_height;
		var->xres_virtual=video_width;
		var->yres_virtual=video_height_virtual;
		var->xoffset=0;
		var->yoffset=0;
		var->grayscale=0;
		var->bits_per_pixel=video_bpp;
		var->nonstd=0;
		var->accel_flags=0;

		var->red.msb_right = 0;
		var->green.msb_right = 0;
		var->blue.msb_right = 0;
		var->transp.msb_right = 0;

		switch(var->bits_per_pixel) {
		case 8:
		    temporal=0;
	    	break;
		case 15:
		    temporal=1;
	    	break;
		case 16:
		    temporal=2;
	    	break;
		case 24:
			temporal=3;
			break;
		case 32:
		    temporal=4;
	    	break;
		default:
		    temporal=0;
		}
	
		var->red.offset= vesafb_color_data[temporal][0][1];
		var->red.length = vesafb_color_data[temporal][0][0];
		var->green.offset = vesafb_color_data[temporal][1][1];
		var->green.length = vesafb_color_data[temporal][1][0];
		var->blue.offset = vesafb_color_data[temporal][2][1];
		var->blue.length = vesafb_color_data[temporal][2][0];
		var->transp.offset = 0;
		var->transp.length = 0;
	
	} else
		*var=fb_display[con].var;
	
    return 0;
}

static void vesafb_set_disp(int con)
{
    struct fb_fix_screeninfo fix;
    struct display *display;
    struct display_switch *sw;
    
    if (con >= 0)
		display = &fb_display[con];
    else
		display = &disp;	/* used during initialization */
    
    vesafb_get_fix(&fix, con, 0);
    
    memset(display, 0, sizeof(struct display));
    display->screen_base = video_vbase;
    display->visual = fix.visual;
    display->type = fix.type;
    display->type_aux = fix.type_aux;
    display->ypanstep = fix.ypanstep;
    display->ywrapstep = fix.ywrapstep;
    display->line_length = fix.line_length;
    display->next_line = fix.line_length;
    display->can_soft_blank = 0;
    display->inverse = inverse;
    vesafb_get_var(&display->var, -1, &fb_info);
    
    switch (video_bpp) {
    case 8:
		sw = &fbcon_cfb8;
		break;
    case 15:
    case 16:
		sw = &fbcon_cfb16;
		display->dispsw_data = fbcon_cmap.cfb16;
	break;
    case 24:
		sw = &fbcon_cfb24;
		display->dispsw_data = fbcon_cmap.cfb24;
		break;
    case 32:
		sw = &fbcon_cfb32;
		display->dispsw_data = fbcon_cmap.cfb32;
		break;
    default:
#ifdef FBCON_HAS_MAC
		sw = &fbcon_mac;
		break;
#else
		sw = &fbcon_dummy;
		return;
#endif
    }
    memcpy(&vesafb_sw, sw, sizeof(struct display_switch));
    if (!ypan) {
		display->scrollmode = SCROLL_YREDRAW;
		vesafb_sw.bmove = fbcon_redraw_bmove;
    }
    display->dispsw = &vesafb_sw;
}

static int vesafb_set_var2(struct fb_var_screeninfo *var, int con,
			      struct fb_info *info,unsigned char tipo) {
    
	struct display *display;
    unsigned short int new_mode;
	unsigned int multiplier=1;    

    char *argv[2];
    char *envp[1];

    envp[0]=NULL;
    argv[0]="/sbin/fbsetmoded";
    argv[1]=NULL;

    if(con>=0)
		display=&fb_display[con];
    else
		display=&disp;
    
    if((var->xres_virtual!=var->xres) ||
       (var->yres_virtual != var->yres) ||
       var->xoffset ||
       ((var->bits_per_pixel !=8)&&(var->bits_per_pixel !=15)&&(var->bits_per_pixel !=16)&&(var->bits_per_pixel !=24)&&(var->bits_per_pixel !=32)) ||
       var->nonstd)
		return -EINVAL;

	if(0==vesafb_find_mode((unsigned short int)var->xres,(unsigned short int)var->yres,(unsigned short int)var->bits_per_pixel,&new_mode)) {
		if(vesafb_got_capabilities == 0)
			printk(KERN_ERR "No graphic modes available.\nBe sure you run the daemon at startup.\n");
		if(vesafb_got_capabilities == 2)
			printk(KERN_ERR "Default graphic mode no available.\nStart your system in other graphic mode.\n");
		return -EINVAL;
	}
	
    if ((var->yoffset)&& !ypan)
		return -EINVAL;
    
    if ((var->activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_TEST)
		return 0;
    
    if(con != currcon)
		return 0;

	if(vesafb_got_capabilities != 1)
		return 0; // if we reached this point and there's no available modes, we are requiring the current mode :)
	
	switch(var->bits_per_pixel) {
	case 8:
		video_visual   = FB_VISUAL_PSEUDOCOLOR;
		video_bpp      = 8;
		multiplier     = 1;
		video_cmap_len = 256;
	break;
	case 15:
		video_visual   = FB_VISUAL_TRUECOLOR;
		video_bpp      = 15;
		multiplier     = 2;
		video_cmap_len = 16;
	break;
	case 16:
		video_visual   = FB_VISUAL_TRUECOLOR;
		video_bpp      = 16;
		multiplier     = 2;
		video_cmap_len = 16;
	break;
	case 24:
		video_visual   = FB_VISUAL_TRUECOLOR;
		video_bpp      = 24;
		multiplier     = 3;
		video_cmap_len = 16;
	break;
	case 32:
		video_visual   = FB_VISUAL_TRUECOLOR;
		video_bpp      = 32;
		multiplier     = 4;
		video_cmap_len = 16;
	break;
	}
	
	video_width = var->xres;
	video_height = var->yres;
	
	video_linelength = multiplier * video_width;
		
    video_height_virtual = video_height;
    
    spin_lock(&vesafb_varlock); // lock the common variables
	
	vesafb_dflags |= 1; // new mode
	vesafb_daemon_command[1]=((unsigned char) new_mode); // new graphic mode
	vesafb_daemon_command[2]=((unsigned char) con); // console number
	
	if (vesafb_dflags & 0x02) { // are there pending palette changes?
		
		unsigned char *puntero;
		int bucle;
		
		vesafb_dflags &= 0xFD; // invalidate all
		puntero = vesafb_palete;
		for(bucle=0;bucle<256;bucle++) {
			*puntero=0;
			puntero+=3;
		}
	}
		
	
	if((vesafb_dflags & 0x04) == 0) { // if there's already a daemon launched, we don't launch another
		vesafb_dflags |= 0x04;
    	call_usermodehelper(argv[0],argv,envp);
	}

	spin_unlock(&vesafb_varlock);

    return 0;
}

static int vesafb_read_proc(char *page,char **start,off_t off,int count,int *eof,void *data) {

	unsigned int bucle;
	unsigned char *entrada,numero,found;
	
    spin_lock_irq(&vesafb_varlock); // lock the common variables

	vesafb_daemon_command[0]=0;
	
	if(vesafb_dflags & 0x01) { // has to change the graphic mode
		vesafb_dflags &= 0xFE; // clear the bit
		vesafb_daemon_command[0]=1;   // command to change only the graphic mode
	}
	
	if (vesafb_dflags & 0x02) { // has to send a palette entry
		entrada=vesafb_palete;
		numero=found=0;
		for (bucle=0;bucle<256;bucle++)
			if ((*entrada) > 127) { // entry changed
				found=1;
				break;
			} else {
				entrada+=3;
				numero++;
			}
		
		if (found) {
			(*entrada) -= 128; // restore the right value
			vesafb_daemon_command[0]+=2; // change palette too
			vesafb_daemon_command[3] = numero;
			vesafb_daemon_command[4] = *(entrada++);
			vesafb_daemon_command[5] = *(entrada++);
			vesafb_daemon_command[6] = *entrada;
		} else
			vesafb_dflags &= 0xFD; // clear the bit
	}	
	
	if(vesafb_daemon_command[0]==0) // if we sent an "end command"
		vesafb_dflags = 0; // all is done
	
	for(bucle=0;bucle<7;bucle++)
		*(page+bucle)=vesafb_daemon_command[bucle];
	
	*eof=1;
	
	spin_unlock_irq(&vesafb_varlock);

	return 7; // seven chars returned
	
}

int vesafb_write_proc(struct file *fichero, const char *buffer, unsigned long count, void *data) {

	static unsigned char status = 0; // waiting for a START command
	static unsigned char cuantity = 0;
	static unsigned char default_mode_found = 0; //if 1, the mode used to boot has been found in the list
	
	unsigned char lbuf[8];
	
	int bucle,size;
	unsigned char *data1;
	unsigned char ctempo1;
	unsigned short int *data2;
	unsigned short int itempo1,itempo2,itempo3;
	
	if(count > 8)
		size = 8;
	else
		size = count;
	
	copy_from_user(lbuf,buffer,size);

	// printk(KERN_INFO "comando: %d   %x %x %x %x %x %x %x %x\n",(int)count,lbuf[0],lbuf[1],lbuf[2],lbuf[3],lbuf[4],lbuf[5],lbuf[6],lbuf[7]);
	
	if(status == 2) // transfer ended. No more commands allowed
		return (count);
	
	data1 = (unsigned char *) buffer;
	data2 = (unsigned short int *) buffer;
	
	switch(status) {
	case 0: // waiting for a START command
		if(size<2)
			return (count);
		
		if((*data1) != 0)
			return (count); // not an START command
		
		data1++;
		cuantity = *data1; // number of entries		
		vesafb_supported_modes = kmalloc(6*(1+(unsigned int) cuantity),GFP_KERNEL);
		
		if(NULL == vesafb_supported_modes) {
			printk(KERN_ERR "Can't assign kernel memory\n");
			return (count);
		}
		
		for(bucle=0;bucle<(3*(1+((unsigned int)cuantity)));bucle++)
			vesafb_supported_modes[bucle]=0; // all to zeroes
		
		status = 1; // waiting for COLOR or MODE commands
	break;
	
	case 1: // waiting for a COLOR or MODE command
	
		switch(*(data1++)) {
		case 1: // COLOR command
			ctempo1 = *(data1++);
			vesafb_color_data[ctempo1][0][0] = *(data1++);
			vesafb_color_data[ctempo1][0][1] = *(data1++);
			vesafb_color_data[ctempo1][1][0] = *(data1++);
			vesafb_color_data[ctempo1][1][1] = *(data1++);
			vesafb_color_data[ctempo1][2][0] = *(data1++);
			vesafb_color_data[ctempo1][2][1] = *(data1++);
			break;
		
		case 2: // MODE command
			ctempo1 = *data1; // entry number
			data2++;
			itempo1=*(data2++); // resx
			itempo2=*(data2++); // resy
			itempo3=*data2; // bpp and mode
			vesafb_supported_modes[3*((unsigned int)ctempo1)] = itempo1;
			vesafb_supported_modes[1+3*((unsigned int)ctempo1)] = itempo2;
			vesafb_supported_modes[2+3*((unsigned int)ctempo1)] = itempo3;
			if ((((int)itempo1) == video_width)&&(((int)itempo2) == video_height)&&(((int)((itempo3>>8)&0x00FF)) == video_bpp))
				default_mode_found = 1;
			break;
		
		case 3: // END command
			status = 2;
			vesafb_got_capabilities = (default_mode_found == 1) ? 1 : 2;
			if(vesafb_got_capabilities == 2)
				printk(KERN_ERR "Default graphic mode not found. Emulating classic VESAFB.\n");
			break;		
		}
	}
	return (count);
	
}

static int vesafb_set_var(struct fb_var_screeninfo *var, int con,
			  struct fb_info *info)
{
    
    int retorno,temporal=0;
    struct display *display;
    struct fb_fix_screeninfo fix;	
    
    allow_set = 1;

    retorno=vesafb_set_var2(var, con, info,0);
    if (retorno)
		return(retorno);

	if(vesafb_got_capabilities != 1)
		return 0;
	
    if ((var->activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_TEST) {
		var->vmode=FB_VMODE_CONUPDATE;
		return 0;
    }
    
    if(con<0)
		return 0;

    display=&fb_display[con];

    display->var=*var;
    display->var.vmode=FB_VMODE_CONUPDATE;
    display->screen_base=video_vbase;
    vesafb_get_fix(&fix,-1,NULL);
    display->visual=fix.visual;
    display->type = fix.type;
    display->type_aux = fix.type_aux;
    display->ypanstep = fix.ypanstep;
    display->ywrapstep = fix.ywrapstep;
    display->line_length = fix.line_length;
    display->next_line=fix.line_length;
    display->can_soft_blank = 1;
    display->inverse = 0;
    
    if(currcon!=con) {
		switch(display->var.bits_per_pixel) {
		case 8:
		    display->visual=FB_VISUAL_PSEUDOCOLOR;
		    display->line_length=var->xres_virtual;
	    	display->next_line=var->xres_virtual;
		    break;
		case 15:
	    	display->visual=FB_VISUAL_TRUECOLOR;
	    	display->line_length=2*(var->xres_virtual);
	    	display->next_line=2*(var->xres_virtual);
	    	break;	
		case 16:
		    display->visual=FB_VISUAL_TRUECOLOR;
	    	display->line_length=2*(var->xres_virtual);
	    	display->next_line=2*(var->xres_virtual);
	    	break;	
		case 24:
		    display->visual=FB_VISUAL_TRUECOLOR;
	    	display->line_length=3*(var->xres_virtual);
	    	display->next_line=3*(var->xres_virtual);
	    	break;
		case 32:
		    display->visual=FB_VISUAL_TRUECOLOR;
	    	display->line_length=4*(var->xres_virtual);
	    	display->next_line=4*(var->xres_virtual);
	    	break;
		}
    }

    display->var.red.msb_right = 0;
    display->var.green.msb_right = 0;
    display->var.blue.msb_right = 0;
    display->var.transp.msb_right = 0;

    switch(display->var.bits_per_pixel) {
    case 8:
		display->dispsw = &fbcon_cfb8;
		break;
    case 15:
		display->dispsw = &fbcon_cfb16;
		display->dispsw_data = fbcon_cmap.cfb16;
		break;
    case 16:
		display->dispsw = &fbcon_cfb16;
		display->dispsw_data = fbcon_cmap.cfb16;
		break;
    case 24:
		display->dispsw = &fbcon_cfb24;
		display->dispsw_data = fbcon_cmap.cfb24;
		break;
    case 32:
		display->dispsw = &fbcon_cfb32;
		display->dispsw_data = fbcon_cmap.cfb32;
		break;
    }

    switch(var->bits_per_pixel) {
    case 8:
		temporal=0;
		break;
    case 15:
		temporal=1;
		break;
    case 16:
		temporal=2;
		break;
    case 24:
		temporal=3;
		break;
    case 32:
		temporal=4;
		break;
    }
    
    display->var.red.offset= vesafb_color_data[temporal][0][1];
    display->var.red.length = vesafb_color_data[temporal][0][0];
    display->var.green.offset = vesafb_color_data[temporal][1][1];
    display->var.green.length = vesafb_color_data[temporal][1][0];
    display->var.blue.offset = vesafb_color_data[temporal][2][1];
    display->var.blue.length = vesafb_color_data[temporal][2][0];
    display->var.transp.offset = 0;
    display->var.transp.length = 0;
    
    if (fb_info.changevar)
		(*fb_info.changevar)(con);

    return 0;
}

static int vesa_getcolreg(unsigned regno, unsigned *red, unsigned *green,
			     unsigned *blue, unsigned *transp,
			     struct fb_info *fb_info)
{
    /*
     *  Read a single color register and split it into colors/transparent.
     *  Return != 0 for invalid regno.
     */
    
    if (regno >= video_cmap_len)
		return 1;
    
    *red   = palette[regno].red;
    *green = palette[regno].green;
    *blue  = palette[regno].blue;
    *transp = 0;
    return 0;
}

static void vesa_setpalette(int regno, unsigned red, unsigned green, unsigned blue)
{
#ifdef i386
	/* without protected mode interface, try VGA registers... */
		
	spin_lock(vesafb_varlock);
		
	if (vesafb_dflags & 0x04) { // if there's a daemon launched
			
		vesafb_dflags |= 0x02; // there's a changed entry
		vesafb_palete[regno*3]   = (((unsigned char)(red >> 10)) & 0x3F) + 128; // bit 7 set
		vesafb_palete[1+regno*3] = ((unsigned char)(green >> 10)) & 0x3F;
		vesafb_palete[2+regno*3] = ((unsigned char)(blue >> 10)) & 0x3F;
		
	} else {
		
		outb_p(regno,       dac_reg);
		outb_p(red   >> 10, dac_val);
		outb_p(green >> 10, dac_val);
		outb_p(blue  >> 10, dac_val);
    }
	
	spin_unlock(vesafb_varlock);
#endif
}

static int vesa_setcolreg(unsigned regno, unsigned red, unsigned green,
			     unsigned blue, unsigned transp,
			     struct fb_info *fb_info)
{
    /*
     *  Set a single color register. The values supplied are
     *  already rounded down to the hardware's capabilities
     *  (according to the entries in the `var' structure). Return
     *  != 0 for invalid regno.
     */
    
    if (regno >= video_cmap_len)
		return 1;
    
    palette[regno].red   = red;
    palette[regno].green = green;
    palette[regno].blue  = blue;
    
    switch (video_bpp) {
    case 8:
		vesa_setpalette(regno,red,green,blue);
	break;
    case 15:
	fbcon_cmap.cfb16[regno] =
	    (((red   & 0xf800) >> (16-(vesafb_color_data[1][0][0]))) << (vesafb_color_data[1][0][1])) |
	    (((green & 0xf800) >> (16-(vesafb_color_data[1][1][0]))) << (vesafb_color_data[1][1][1])) |
	    (((blue  & 0xf800) >> (16-(vesafb_color_data[1][2][0]))) << (vesafb_color_data[1][2][1]));
	break;
    case 16:
	fbcon_cmap.cfb16[regno] =
	    (((red   & 0xf800) >> (16-(vesafb_color_data[2][0][0]))) << (vesafb_color_data[2][0][1])) |
	    (((green & 0xfc00) >> (16-(vesafb_color_data[2][1][0]))) << (vesafb_color_data[2][1][1])) |
	    (((blue  & 0xf800) >> (16-(vesafb_color_data[2][2][0]))) << (vesafb_color_data[2][2][1]));	    
	break;
    case 24:
	red   >>= 8;
	green >>= 8;
	blue  >>= 8;
	fbcon_cmap.cfb24[regno] =
	    (red   << (vesafb_color_data[3][0][1]))   |
	    (green << (vesafb_color_data[3][1][1]))   |
	    (blue  << (vesafb_color_data[3][2][1]));
	break;
    case 32:
	red   >>= 8;
	green >>= 8;
	blue  >>= 8;
	fbcon_cmap.cfb32[regno] =
	    (red   << (vesafb_color_data[4][0][1]))   |
	    (green << (vesafb_color_data[4][1][1]))   |
	    (blue  << (vesafb_color_data[4][2][1]));
	break;
    }
    return 0;
}

static void do_install_cmap(int con, struct fb_info *info)
{
    if (con != currcon)
		return;
	
    if (fb_display[con].cmap.len)
		fb_set_cmap(&fb_display[con].cmap, 1, vesa_setcolreg, info);
    else
		fb_set_cmap(fb_default_cmap(video_cmap_len), 1, vesa_setcolreg,info);
}

static int vesafb_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			      struct fb_info *info)
{
	int vesafb_cmap_len = (fb_display[con].visual == FB_VISUAL_PSEUDOCOLOR) ? 256 : 16;
		
    if (con == currcon) /* current console? */
		return fb_get_cmap(cmap, kspc, vesa_getcolreg, info);
    else if (fb_display[con].cmap.len) /* non default colormap? */
		fb_copy_cmap(&fb_display[con].cmap, cmap, kspc ? 0 : 2);
    else
		fb_copy_cmap(fb_default_cmap(vesafb_cmap_len),cmap, kspc ? 0 : 2);
    return 0;
}

static int vesafb_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			      struct fb_info *info)
{
    int err;

	int vesafb_cmap_len = (fb_display[con].visual == FB_VISUAL_PSEUDOCOLOR) ? 256 : 16;
	    
    if (!fb_display[con].cmap.len) {	/* no colormap allocated? */
		err = fb_alloc_cmap(&fb_display[con].cmap,vesafb_cmap_len,0);
		if (err)
	    	return err;
    }
	
    if (con == currcon)			/* current console? */
		return fb_set_cmap(cmap, kspc, vesa_setcolreg, info);
    else
		fb_copy_cmap(cmap, &fb_display[con].cmap, kspc ? 0 : 1);
	
    return 0;
}

static struct fb_ops vesafb_ops = {
    owner:		THIS_MODULE,
    fb_get_fix:	        vesafb_get_fix,
    fb_get_var:	        vesafb_get_var,
    fb_set_var:	        vesafb_set_var,
    fb_get_cmap:		vesafb_get_cmap,
    fb_set_cmap:		vesafb_set_cmap,
    fb_pan_display:		vesafb_pan_display,
};

int __init vesafb_setup(char *options)
{
    char *this_opt;
    
    fb_info.fontname[0] = '\0';

    if (!options || !*options)
		return 0;

//    for(this_opt=strtok(options,","); this_opt; this_opt=strtok(NULL,",")) { // changed
	while ((this_opt = strsep(&options, ",")) != NULL) {
		if (!*this_opt) continue;
	
		if (! strcmp(this_opt, "inverse"))
	    	inverse=1;
		else if (! strcmp(this_opt, "redraw"))
		    ypan=0;
		else if (! strcmp(this_opt, "ypan"))
	    	ypan=1;
		else if (! strcmp(this_opt, "ywrap"))
		    ypan=2;
		else if (! strcmp(this_opt, "vgapal"))
	    	pmi_setpal=0;
		else if (! strcmp(this_opt, "pmipal"))
		    pmi_setpal=1;
		else if (! strcmp(this_opt, "mtrr"))
	    	mtrr=1;
		else if (!strncmp(this_opt, "font:", 5))
		    strcpy(fb_info.fontname, this_opt+5);
    }
	pmi_setpal=0;
	ypan=0;
    return 0;
}

static int vesafb_switch(int con, struct fb_info *info) {
    
	if(con<0)
		return 0;
	
    /* Do we have to save the colormap? */
    if ((currcon >=0)&&(fb_display[currcon].cmap.len))
		fb_get_cmap(&fb_display[currcon].cmap, 1, vesa_getcolreg,info);
    
	if((allow_set)&&(currcon != con)) { // only allow to change the graphic mode if we already has made it.		
		currcon = con;
		vesafb_set_var2(&fb_display[con].var,con,info,1); // this avoids a hang while booting
	}
	
	currcon = con;
	do_install_cmap(con, info);

    return 0;
}

/* 0 unblank, 1 blank, 2 no vsync, 3 no hsync, 4 off */

static void vesafb_blank(int blank, struct fb_info *info)
{
    /* Not supported */
}

int __init vesafb_init(void)
{
    int i,j,indice,bucle;

	vesafb_daemon_command[0]=0; // invalid command
	
	vesafb_fichero_proc = create_proc_entry("supervesafb",0600,NULL);//,vesafb_read_proc,NULL);
	
	if(vesafb_fichero_proc==NULL) {
		printk(KERN_ERR "SuperVESAfb: abort, cannot create /proc file");
		return -ENXIO; // we can't initialize the /proc file
	}

	vesafb_fichero_proc->read_proc  = vesafb_read_proc;
	vesafb_fichero_proc->write_proc = vesafb_write_proc;
	
	vesafb_specs_proc = create_proc_read_entry("vesafbmodes",0444,NULL,vesafb_read_specs,NULL);

	if(vesafb_specs_proc==NULL) {
		printk(KERN_ERR "SuperVESAfb: abort, cannot create /proc file");
		return -ENXIO; // we can't initialize the /proc file
	}
	
    if (screen_info.orig_video_isVGA != VIDEO_TYPE_VLFB) {
		remove_proc_entry("supervesafb",NULL);
		remove_proc_entry("vesafbmodes",NULL);
		return -ENXIO;
	}
	
    video_base          = screen_info.lfb_base;
    video_bpp           = screen_info.lfb_depth;
    video_width         = screen_info.lfb_width;
    video_height        = screen_info.lfb_height;
    video_linelength    = screen_info.lfb_linelength;
    video_size          = screen_info.lfb_width * screen_info.lfb_height * video_bpp;
    video_visual = (video_bpp == 8) ? FB_VISUAL_PSEUDOCOLOR : FB_VISUAL_TRUECOLOR;
    
#ifndef __i386__
	screen_info.vesapm_seg = 0;
#endif
	
    if (!request_mem_region(video_base, video_size, "SuperVESAfb")) {
		printk(KERN_WARNING
	       "SuperVESAfb: abort, cannot reserve video memory at 0x%lx\n",
	       video_base);
		//remove_proc_entry("supervesafb",NULL); // changed
		//remove_proc_entry("vesafbmodes",NULL); // changed
		//return -EBUSY; // changed
		/* We cannot make this fatal. Sometimes this comes from magic
        spaces our resource handlers simply don't know about */		
    }
    
    video_vbase = ioremap(video_base, video_size);
    if (!video_vbase) {
		release_mem_region(video_base, video_size);
		printk(KERN_ERR
	       "SuperVESAfb: abort, cannot ioremap video memory 0x%x @ 0x%lx\n",
	       video_size, video_base);
		remove_proc_entry("supervesafb",NULL);
		remove_proc_entry("vesafbmodes",NULL);
		return -EIO;
    }
    
    printk(KERN_INFO "SuperVESAfb: framebuffer at 0x%lx, mapped to 0x%p, size %dk\n",
	   video_base, video_vbase, video_size/1024);
    printk(KERN_INFO "SuperVESAfb: mode is %dx%dx%d, linelength=%d, pages=%d\n",
	   video_width, video_height, video_bpp, video_linelength, screen_info.pages);
    
    if (screen_info.vesapm_seg) {
		printk(KERN_INFO "SuperVESAfb: protected mode interface info at %04x:%04x\n",
	       screen_info.vesapm_seg,screen_info.vesapm_off);
    }
    
    if (screen_info.vesapm_seg < 0xc000)
		ypan = pmi_setpal = 0; /* not available or some DOS TSR ... */
    
    if (ypan || pmi_setpal) {
		pmi_base  = (unsigned short*)bus_to_virt(((unsigned long)screen_info.vesapm_seg << 4) + screen_info.vesapm_off);
		pmi_start = (void*)((char*)pmi_base + pmi_base[1]);
		pmi_pal   = (void*)((char*)pmi_base + pmi_base[2]);
		printk(KERN_INFO "SuperVESAfb: pmi: set display start = %p, set palette = %p\n",pmi_start,pmi_pal);
		if (pmi_base[3]) {
		    printk(KERN_INFO "SuperVESAfb: pmi: ports = ");
		    for (i = pmi_base[3]/2; pmi_base[i] != 0xffff; i++)
			printk("%x ",pmi_base[i]);
	    	printk("\n");
	    	if (pmi_base[i] != 0xffff) {
				/*
				 * memory areas not supported (yet?)
				 *
				 * Rules are: we have to set up a descriptor for the requested
				 * memory area and pass it in the ES register to the BIOS function.
				 */
				printk(KERN_INFO "SuperVESAfb: can't handle memory requests, pmi disabled\n");
				ypan = pmi_setpal = 0;
	    	}
		}
    }
    
    vesafb_defined.xres=video_width;
    vesafb_defined.yres=video_height;
    vesafb_defined.xres_virtual=video_width;
    vesafb_defined.yres_virtual=video_size / video_linelength;
    vesafb_defined.bits_per_pixel=video_bpp;
    
    if (ypan && vesafb_defined.yres_virtual > video_height) {
		printk(KERN_INFO "SuperVESAfb: scrolling: %s using protected mode interface, yres_virtual=%d\n",
	       (ypan > 1) ? "ywrap" : "ypan",vesafb_defined.yres_virtual);
    } else {
		printk(KERN_INFO "SuperVESAfb: scrolling: redraw\n");
		vesafb_defined.yres_virtual = video_height;
		ypan = 0;
    }
    video_height_virtual = vesafb_defined.yres_virtual;
    
    /* some dummy values for timing to make fbset happy */
    vesafb_defined.pixclock     = 10000000 / video_width * 1000 / video_height;
    vesafb_defined.left_margin  = (video_width / 8) & 0xf8;
    vesafb_defined.right_margin = 32;
    vesafb_defined.upper_margin = 16;
    vesafb_defined.lower_margin = 4;
    vesafb_defined.hsync_len    = (video_width / 8) & 0xf8;
    vesafb_defined.vsync_len    = 4;
    
    if (video_bpp > 8) {
		vesafb_defined.red.offset    = screen_info.red_pos;
		vesafb_defined.red.length    = screen_info.red_size;
		vesafb_defined.green.offset  = screen_info.green_pos;
		vesafb_defined.green.length  = screen_info.green_size;
		vesafb_defined.blue.offset   = screen_info.blue_pos;
		vesafb_defined.blue.length   = screen_info.blue_size;
		vesafb_defined.transp.offset = screen_info.rsvd_pos;
		vesafb_defined.transp.length = screen_info.rsvd_size;
		printk(KERN_INFO "SuperVESAfb: directcolor: "
	       "size=%d:%d:%d:%d, shift=%d:%d:%d:%d\n",
	       screen_info.rsvd_size,
	       screen_info.red_size,
	       screen_info.green_size,
	       screen_info.blue_size,
	       screen_info.rsvd_pos,
	       screen_info.red_pos,
	       screen_info.green_pos,
	       screen_info.blue_pos);
		video_cmap_len = 16;
    } else {
		vesafb_defined.red.length   = 6;
        vesafb_defined.green.length = 6;
		vesafb_defined.blue.length  = 6;
		for(i = 0; i < 16; i++) {
	    	j = color_table[i];
	    	palette[i].red   = default_red[j];
	    	palette[i].green = default_grn[j];
	    	palette[i].blue  = default_blu[j];
		}
		video_cmap_len = 256;
    }
    
	for(bucle=0;bucle<6;bucle++)
		vesafb_supported_modes2[bucle]=0; // empty the alternative list
	
	vesafb_supported_modes2[0]=((unsigned short int)video_width);
	vesafb_supported_modes2[1]=((unsigned short int)video_height);
	vesafb_supported_modes2[2]=((unsigned short int)((video_bpp<<8)&0xFF00));
	
	indice=0;
	
	switch(video_bpp) {
	case 8:
		indice=0;
	break;
	case 15:
		indice=1;
	break;
	case 16:
		indice=2;
	break;
	case 24:
		indice=3;
	break;
	case 32:
		indice=4;
	break;
	}
	
	vesafb_color_data[indice][0][0]=vesafb_defined.red.length;
	vesafb_color_data[indice][0][1]=vesafb_defined.red.offset;
	vesafb_color_data[indice][1][0]=vesafb_defined.green.length;
	vesafb_color_data[indice][1][1]=vesafb_defined.green.offset;
	vesafb_color_data[indice][2][0]=vesafb_defined.blue.length;
	vesafb_color_data[indice][2][1]=vesafb_defined.blue.offset;
	
    /* request failure does not faze us, as vgacon probably has this
     * region already (FIXME) */
    request_region(0x3c0, 32, "SuperVESAfb");
    
	
	if (mtrr) {
		int temp_size = video_size;
		/* Find the largest power-of-two */
		while (temp_size & (temp_size - 1))
                	temp_size &= (temp_size - 1);
                        
        /* Try and find a power of two to add */
		while (temp_size && mtrr_add(video_base, temp_size, MTRR_TYPE_WRCOMB, 1)==-EINVAL) {
			temp_size >>= 1;
		}
	}
     
    strcpy(fb_info.modename, "VESA VGA");
    fb_info.changevar  = NULL;
    fb_info.node       = -1;
    fb_info.fbops      = &vesafb_ops;
    fb_info.disp       = &disp;
    fb_info.switch_con = &vesafb_switch;
    fb_info.updatevar  = &vesafb_update_var;
    fb_info.blank      = &vesafb_blank;
    fb_info.flags      = FBINFO_FLAG_DEFAULT;
    vesafb_set_disp(-1);
    
    if (register_framebuffer(&fb_info)<0) {
		remove_proc_entry("supervesafb",NULL);
		remove_proc_entry("vesafbmodes",NULL);		
		return -EINVAL;
	}
    
    printk(KERN_INFO "fb%d: %s frame buffer device\n",
	       GET_FB_IDX(fb_info.node), fb_info.modename);
	
    return 0;
}

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-basic-offset: 4
 * End:
 */

MODULE_LICENSE("GPL");
